@@ -21,11 +21,16 @@ except ImportError:
     pass
 
 from xml2rfc import strings, log
-from xml2rfc.util.date import extract_date, format_date, get_expiry_date
+from xml2rfc.util.date import extract_date, augment_date, format_date, get_expiry_date
 from xml2rfc.util.name import short_author_ascii_name_parts, full_author_name_expansion, short_author_name_parts
-from xml2rfc.util.unicode import punctuation, unicode_replacements, unicode_content_tags, downcode
+from xml2rfc.util.unicode import ( punctuation, unicode_replacements, unicode_content_tags, bare_unicode_tags,
+    unicode_attributes, downcode, downcode_punctuation)
 from xml2rfc.utils import namespaces, find_duplicate_ids
 
+default_silenced_messages = [
+#    ".*[Pp]ostal address",
+]
+
 default_options = Values(defaults={
         'accept_prepped': None,
         'add_xinclude': None,
@@ -41,6 +46,7 @@ default_options = Values(defaults={
         'dtd': None,
         'expand': False,
         'external_css': False,
+        'external_js': True,            # Change to False in 3.x
         'filename': None,
         'first_page_author_org': True,
         'html': False,
@@ -51,7 +57,7 @@ default_options = Values(defaults={
         'indent': 2,
         'info': False,
         'legacy': True,
-        'legacy_date_format': False,
+        'legacy_date_format': False,    # Change to True in 3.x
         'legacy_list_symbols': False,
         'list_symbols': ('*', '-', 'o', '+'),
         'metadata_js_url': 'metadata.min.js',
@@ -71,6 +77,7 @@ default_options = Values(defaults={
         'rfc': None,
         'rfc_base_url': 'https://www.rfc-editor.org/rfc/',
         'rfc_reference_base_url': 'https://rfc-editor.org/rfc/',
+        'silence': default_silenced_messages,
         'strict': False,
         'text': True,
         'utf8': False,
@@ -1247,8 +1254,11 @@ class BaseRfcWriter:
                 ref_counter += 1
                 title = ref.find("front/title")
                 if title != None:
-                    self._indexRef(ref_counter, title=title.text, anchor=ref.attrib["anchor"])
-
+                    if 'anchor' in ref.attrib:
+                        self._indexRef(ref_counter, title=title.text, anchor=ref.attrib["anchor"])
+                    else:
+                        raise RfcWriterError("Reference is missing an anchor: %s" % lxml.etree.tostring(ref))
+                        
         # Appendix sections
         back = self.r.find('back')
         if back is not None:
@@ -1628,11 +1638,20 @@ def get_inline_tags():
                 tags.add(name)
     return tags
 
+def get_xref_tags():
+    "Get tags that can occur as direct children of <xref>"
+    tags = set()
+    refs = v3_schema.xpath("/x:grammar/x:define/x:element[@name='xref']//x:ref", namespaces=namespaces)
+    tags = set([ c.get('name') for c in refs ])
+    return tags
+
+
 meta_tags   = get_meta_tags() - deprecated_element_tags
 text_tags   = get_text_tags() - deprecated_element_tags
 inline_tags = get_inline_tags() - deprecated_element_tags
-block_tags  = get_element_tags() - text_tags - deprecated_element_tags
+block_tags  = get_element_tags() - inline_tags - deprecated_element_tags
 mixed_tags  = get_text_or_block_tags() - deprecated_element_tags
+xref_tags   = get_xref_tags()
 
 # --------------------------------------------------------------------------------------------------
 
@@ -1653,9 +1672,9 @@ class BaseV3Writer(object):
         self.text_tags = set(text_tags)
         self.inline_tags = set(inline_tags)
         self.mixed_tags = set(mixed_tags)
-        self.refname_mapping = dict( (e.get('anchor'), e.get('anchor')) for e in self.root.xpath('.//reference') )
-        self.refname_mapping.update(dict( (e.get('target'), e.get('to')) for e in self.root.xpath('.//displayreference') ))
-        self.refname_mapping.update(dict( (e.get('anchor'), e.get('anchor')) for e in self.root.xpath('.//referencegroup') ))
+        self.xref_tags = set(xref_tags)
+        self.attribute_defaults = self.get_all_attribute_defaults()
+        #
         #
         self.errors = []
 #         if options.debug:
@@ -1714,7 +1733,7 @@ class BaseV3Writer(object):
     def silenced(self, e, text):
         text = text.strip()
         pis = self.get_relevant_pis(e)
-        silenced = filter(None, [ pi.get('silence') for pi in pis ])
+        silenced = filter(None, self.options.silence[:] + [ pi.get('silence') for pi in pis ])
         for s in silenced:
             for label in ['Note: ', 'Warning: ']:
                 if s.startswith(label):
@@ -1733,9 +1752,9 @@ class BaseV3Writer(object):
             if not self.silenced(e, text):
                 self.log(self.msg(e, label, text))
 
-    def warn(self, e, text):
-        if not self.silenced(e, text):
-            self.log(self.msg(e, 'Warning:', text))
+    def warn(self, e, text, label='Warning:'):
+        if self.options.verbose or not self.silenced(e, text):
+            self.log(self.msg(e, label, text))
 
     def err(self, e, text, trace=False):
         msg = self.msg(e, 'Error:', text)
@@ -1750,6 +1769,31 @@ class BaseV3Writer(object):
         self.errors.append(msg)
         raise RfcWriterError(msg)
 
+    def get_refname_mapping(self):
+        reflist = self.root.xpath('.//references/reference|.//references/referencegroup')
+        if self.root.get('symRefs', 'true') == 'true':
+            refname_mapping = dict( (e.get('anchor'), e.get('anchor')) for e in reflist )
+        else:
+            refname_mapping = (dict( (e.get('anchor'), str(i+1)) for i,e in enumerate(reflist) ))
+        refname_mapping.update(dict( (e.get('target'), e.get('to')) for e in self.root.xpath('.//displayreference') ))
+        return refname_mapping
+
+    def get_all_attribute_defaults(self):
+        defaults = {}
+        elements = self.schema.xpath("/x:grammar/x:define/x:element", namespaces=namespaces)
+        for tag in [ e.get('name') for e in elements ]:
+            attr = self.schema.xpath("/x:grammar/x:define/x:element[@name='%s']//x:attribute" % tag, namespaces=namespaces)
+            defaults[tag] = dict( (a.get('name'), a.get("{%s}defaultValue"%namespaces['a'], None)) for a in attr )
+        return defaults
+
+    def check_refs_numbered(self):
+        # see if references should be numbered.  This is True unless the last
+        # top-level section of <middle/> had numbered='false'.
+        if not hasattr(self, '_refs_numbered'):
+            last_middle_section = ([None, ] + list(self.root.find('middle').iterchildren('section')))[-1]
+            self._refs_numbered = last_middle_section.get('numbered', 'true') == 'true'
+        return self._refs_numbered
+
     # methods operating on the xml tree
 
     def get_element_from_id(self, id):
@@ -1796,8 +1840,9 @@ class BaseV3Writer(object):
 
     def page_top_right(self):
         date = self.root.find('./front/date')
-        parts = extract_date(date, self.options.date)
-        text = format_date(parts[0], parts[1], None, legacy=True)
+        year, month, day = extract_date(date, self.options.date)
+        year, month, day = augment_date(year, month, day, self.options.date)
+        text = format_date(year, month, None, legacy=True)
         return text
 
     def page_bottom_left(self):
@@ -1825,6 +1870,7 @@ class BaseV3Writer(object):
 
     def downcode_punctuation(self):
         self.downcode(replacements=punctuation)
+        self.downcode_reference_punctuation()
 
     def downcode(self, replacements=unicode_replacements):
         """
@@ -1841,19 +1887,38 @@ class BaseV3Writer(object):
                         e.text.encode('ascii')
                     except UnicodeEncodeError:
                         e.text = downcode(e.text, replacements=replacements)
+                elif e.tag in bare_unicode_tags:
+                    pass
+                elif not e.get('ascii'):
+                    try:
+                        e.text.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.text = downcode(e.text, replacements=replacements)
             if e.tail:
                 if not e.getparent().tag in unicode_content_tags:
                     try:
                         e.tail.encode('ascii')
                     except UnicodeEncodeError:
                         e.tail = downcode(e.tail, replacements=replacements)
-            if e.tag != 'author':
-                for key in e.attrib.keys():
+
+    def downcode_attributes(self, replacements=unicode_replacements):
+        for e in self.tree.iter():
+            for key in e.attrib.keys():
+                if not (e.tag, key) in unicode_attributes:
                     try:
                         e.get(key).encode('ascii')
                     except UnicodeEncodeError:
                         e.set(key, downcode(e.get(key), replacements=replacements))
 
+    def downcode_reference_punctuation(self):
+        for r in self.tree.xpath('.//references'):
+            for e in r.iter():
+                for key in e.attrib.keys():
+                    try:
+                        e.get(key).encode('ascii')
+                    except UnicodeEncodeError:
+                        e.set(key, downcode_punctuation(e.get(key)))
+
     def pretty_print_prep(self, e, p):
         ind = self.options.indent
         ## The actual printing is done in self.write()
